Package: HDF5Array
Title: HDF5 datasets as array-like objects in R
Description: The HDF5Array package is an HDF5 backend for DelayedArray objects.
	It implements the HDF5Array, H5SparseMatrix, H5ADMatrix, and
	TENxMatrix classes, 4 convenient and memory-efficient array-like
	containers for representing and manipulating either: (1) a conventional
	(a.k.a. dense) HDF5 dataset, (2) an HDF5 sparse matrix (stored in
	CSR/CSC/Yale format), (3) the central matrix of an h5ad file (or any
	matrix in the /layers group), or (4) a 10x Genomics sparse matrix.
	All these containers are DelayedArray extensions and thus support all
	operations (delayed or block-processed) supported by DelayedArray
	objects.
biocViews: Infrastructure, DataRepresentation, DataImport, Sequencing,
        RNASeq, Coverage, Annotation, GenomeAnnotation, SingleCell,
        ImmunoOncology
URL: https://bioconductor.org/packages/HDF5Array
BugReports: https://github.com/Bioconductor/HDF5Array/issues
Version: 1.40.0
License: Artistic-2.0
Encoding: UTF-8
Authors@R: person("Hervé", "Pagès", role=c("aut", "cre"),
		  email="hpages.on.github@gmail.com",
		  comment=c(ORCID="0009-0002-8272-4522"))
Depends: R (>= 3.4), methods, SparseArray (>= 1.7.5), DelayedArray (>=
        0.33.5), h5mread (>= 1.3.3)
Imports: utils, stats, tools, Matrix, BiocGenerics (>= 0.51.2),
        S4Vectors, IRanges, S4Arrays (>= 1.1.1), rhdf5
Suggests: BiocParallel, GenomicRanges, SummarizedExperiment (>=
        1.15.1), h5vcData, ExperimentHub, TENxBrainData, zellkonverter,
        GenomicFeatures, SingleCellExperiment, DelayedMatrixStats,
        genefilter, RSpectra, RUnit, knitr, rmarkdown, BiocStyle
VignetteBuilder: knitr
Collate: utils.R h5utils.R HDF5ArraySeed-class.R HDF5Array-class.R
        ReshapedHDF5ArraySeed-class.R ReshapedHDF5Array-class.R
        dump-management.R writeHDF5Array.R
        saveHDF5SummarizedExperiment.R H5SparseMatrixSeed-class.R
        H5SparseMatrix-class.R H5ADMatrixSeed-class.R
        H5ADMatrix-class.R TENxMatrixSeed-class.R TENxMatrix-class.R
        writeTENxMatrix.R zzz.R
git_url: https://git.bioconductor.org/packages/HDF5Array
git_branch: RELEASE_3_23
git_last_commit: 6539afc
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 22:45:03 UTC; biocbuild
Author: Hervé Pagès [aut, cre] (ORCID: <https://orcid.org/0009-0002-8272-4522>)
Maintainer: Hervé Pagès <hpages.on.github@gmail.com>
Built: R 4.6.1; ; 2026-07-16 11:55:14 UTC; unix
